//
//  featureProcessing.cpp
//  featMatch
//
//  Created by Abe on 25/06/15.
//  Copyright (c) 2015 Abe. All rights reserved.
//

#include "featureProcessing.h"

FeatureProcessing::FeatureProcessing(){

};

void FeatureProcessing::loadKpDsc(std::string pathToFile, std::vector<cv::KeyPoint> &outputKeypoints, cv::Mat &outputDescriptors){
    cv::FileStorage fs(pathToFile, cv::FileStorage::READ);
    cv::FileNode keypoints = fs["keypoints"];
    cv::FileNode descriptors = fs["descriptors"];
    read(keypoints, outputKeypoints);
    read(descriptors, outputDescriptors);
    fs.release();
};

void FeatureProcessing::saveKpDsc(std::string pathToFile, std::vector<cv::KeyPoint> &inputKeypoints, cv::Mat &inputDescriptors){
    cv::FileStorage fs(pathToFile, cv::FileStorage::WRITE);
    write(fs, "keypoints", inputKeypoints);
    write(fs, "descriptors", inputDescriptors);
    fs.release();
};

void FeatureProcessing::detect(std::string method, cv::Mat inputImage, std::vector<cv::KeyPoint> &outputKeypoints){
    cv::Ptr<cv::FeatureDetector> featureDetector = cv::FeatureDetector::create(method);
    featureDetector->detect(inputImage, outputKeypoints);
};

void FeatureProcessing::extract(std::string method, cv::Mat inputImage, std::vector<cv::KeyPoint> &inputKeypoints, cv::Mat &outputDescriptors){
    cv::Ptr<cv::DescriptorExtractor> featureExtractor = cv::DescriptorExtractor::create(method);
    featureExtractor->compute(inputImage, inputKeypoints, outputDescriptors);

};

void FeatureProcessing::match(bool brute, cv::Mat &input1, cv::Mat &input2, std::vector<cv::DMatch> &outputMatches){

    std::vector<cv::DMatch> initMatches;

    if (brute) {
        //cv::BFMatcher matcher(NORM_L2);
        cv::BFMatcher matcher(cv::NORM_HAMMING);
        matcher.match( input1, input2, initMatches );

    } else {
        cv::FlannBasedMatcher matcher;
        if(input1.type()!=CV_32F) { input1.convertTo(input1, CV_32F); }
        if(input2.type()!=CV_32F) { input2.convertTo(input2, CV_32F); }
        matcher.match( input1, input2, initMatches );
    }


    ////// pro FAST min_dist=90
    double max_dist = 0; double min_dist = 100;

    //-- Quick calculation of max and min distances between keypoints
    for( int i = 0; i < input1.rows; i++ )
    { double dist = initMatches[i].distance;
        if( dist < min_dist ) min_dist = dist;
        if( dist > max_dist ) max_dist = dist;
    }

    /*if (DEBUG) {
        printf("-- Max dist : %f \n", max_dist );
        printf("-- Min dist : %f \n", min_dist );
    }*/

    //-- Take only "good" matches (i.e. whose distance is less than 3*min_dist )

    for( int i = 0; i < input1.rows; i++ ){
        if( initMatches[i].distance < 3*min_dist ){
            outputMatches.push_back( initMatches[i]);
        }
    }


};

void FeatureProcessing::drawKeypoints(cv::Mat img, std::vector<cv::KeyPoint> kp, std::string wnd){

    cv::Mat out;
    cv::Scalar drawColor = cv::Scalar(100, 255, 150);

    cv::drawKeypoints(img, kp, out, drawColor, cv::DrawMatchesFlags::DEFAULT);
    printImage(wnd, out);


};

void FeatureProcessing::drawMatches(cv::Mat img1, std::vector<cv::KeyPoint> key1, cv::Mat img2, std::vector<cv::KeyPoint> key2, std::vector<cv::DMatch> &matches){

    cv::Mat out;
    cv::drawMatches( img1, key1, img2, key2, matches, out, cv::Scalar::all(-1), cv::Scalar::all(-1), std::vector<char>(), cv::DrawMatchesFlags::NOT_DRAW_SINGLE_POINTS );
    printImage( "Good Matches & Object detection", out );

};

void FeatureProcessing::computeTranformationMatrix(std::vector<cv::DMatch> matches, std::vector<cv::KeyPoint> key1, std::vector<cv::KeyPoint> key2, cv::Mat &matrix){

    std::vector<cv::Point2f> obj;
    std::vector<cv::Point2f> scene;

    for( int i = 0; i < matches.size(); i++ )
    {
        //-- Get the keypoints from the good matches
        obj.push_back( key1[ matches[i].queryIdx ].pt );
        scene.push_back( key2[ matches[i].trainIdx ].pt );
    }

    if (obj.size()>3) {
        matrix = cv::findHomography( obj, scene, CV_RANSAC );
    }

};

void FeatureProcessing::drawPerspectiveImage(cv::Mat img, cv::Mat matrix){

    //-- Get the corners from the image_1 ( the object to be "detected" )
    std::vector<cv::Point2f> obj_corners(4);
    obj_corners[0] = cvPoint(0,0);
    obj_corners[1] = cvPoint( img.cols, 0 );
    obj_corners[2] = cvPoint( img.cols, img.rows );
    obj_corners[3] = cvPoint( 0, img.rows );
    std::vector<cv::Point2f> scene_corners(4);


    perspectiveTransform( obj_corners, scene_corners, matrix);

    cv::Mat out;
    cv::warpPerspective(img, out, matrix, out.size());
    printImage("warped", out);

    //-- Draw lines between the corners (the mapped object in the scene - image_2 )
    cv::line( img, scene_corners[0], scene_corners[1], cv::Scalar(0, 255, 0), 4 );
    cv::line( img, scene_corners[1], scene_corners[2], cv::Scalar( 0, 255, 0), 4 );
    cv::line( img, scene_corners[2], scene_corners[3], cv::Scalar( 0, 255, 0), 4 );
    cv::line( img, scene_corners[3], scene_corners[0], cv::Scalar( 0, 255, 0), 4 );

    printImage("Perspective", img);

};

void FeatureProcessing::run(std::string pathToFile, cv::Mat inputImage, cv::Mat &matrix){
    /// Prepare scene descriptors and keypoints
    std::vector<cv::KeyPoint> keyScn;
    cv::Mat descScn;
    loadKpDsc(pathToFile, keyScn, descScn);

    std::vector<cv::KeyPoint> keyObj;
    cv::Mat descObj;
    detect("FAST", inputImage, keyObj);
    extract("BRISK", inputImage, keyObj, descObj);

    /// Match and compute transformation matrix
    std::vector< cv::DMatch > matches;
    match(false, descObj, descScn, matches);
    computeTranformationMatrix(matches, keyObj, keyScn, matrix);

};


void FeatureProcessing::run(cv::Mat inputImage, cv::Mat inputImageScene, cv::Mat &matrix){
    /// Prepare scene descriptors and keypoints
    std::vector<cv::KeyPoint> keyScn;
    cv::Mat descScn;
    std::vector<cv::KeyPoint> keyObj;
    cv::Mat descObj;

    detect("FAST", inputImage, keyObj);
    detect("FAST", inputImageScene, keyScn);
    extract("BRISK", inputImage, keyObj, descObj);
    extract("BRISK", inputImageScene, keyScn, descScn);

    /// Match and compute transformation matrix
    std::vector< cv::DMatch > matches;
    match(false, descObj, descScn, matches);
    computeTranformationMatrix(matches, keyObj, keyScn, matrix);

};


void FeatureProcessing::run(std::string pathToFile, cv::Mat inputImage, cv::Mat &matrix, std::string methodDet, std::string methodExt, bool brute){

    /// Prepare scene descriptors and keypoints
    std::vector<cv::KeyPoint> keyScn;
    cv::Mat descScn;
    loadKpDsc(pathToFile, keyScn, descScn);

    std::vector<cv::KeyPoint> keyObj;
    cv::Mat descObj;
    detect(methodDet, inputImage, keyObj);
    extract(methodExt, inputImage, keyObj, descObj);

    /// Match and compute transformation matrix
    std::vector< cv::DMatch > matches;
    match(brute, descObj, descScn, matches);
    computeTranformationMatrix(matches, keyObj, keyScn, matrix);


};

void FeatureProcessing::printImage(std::string windowName, cv::Mat printedImage){
    cv::imshow(windowName, printedImage);
};
